"""
Module providing helpers to manipulate jobs
"""

import os
import shutil
import quixote
from quixote.build import GeneratorType
import panza.build
from panza.config import get_config
from textwrap import dedent
from subprocess import Popen, run, DEVNULL
import tempfile


class WorkspaceCreationError(Exception):
    """
    Exception class representing an error related to creating a workspace for a job
    """

    def __init__(self, cause: Exception):
        self.__cause__ = cause

    def __str__(self):
        return f"cannot create workspace: {self.__cause__.strerror}: {self.__cause__.filename}"


class EnvironmentBuildError(Exception):
    """
    Exception class representing an error related to the construction of an environment
    """

    def __init__(self, cause: Exception):
        self.__cause__ = cause

    def __str__(self):
        return f"cannot build environment: {str(self.__cause__)}"


class ExecutableJobHandle:
    """
    Class representing a job that is ready to be executed
    """

    def __init__(self, root_dir: str, job_info, blueprint: quixote.Blueprint):
        self.root_dir = root_dir
        self.job_info = job_info
        self.blueprint = blueprint

    def run(self):
        """
        Execute the job
        """
        image_name = f"test-{self.job_info['module_id']}-{self.job_info['activity_id']}"
        volumes = [(self.root_dir, "/moulinette")]

        volumes = map(lambda vol: f"-v {vol[0]}:{vol[1]}", volumes)
        os.system(f"docker run --rm -ti {' '.join(volumes)} {image_name} panza_executor.py")

    def cleanup(self):
        """
        Cleanup the job's working directory
        """
        shutil.rmtree(self.root_dir)


class DataAwaitingJobHandle:
    """
    Class representing a job whose environment is already built, and needs to fetch data to operate on
    """

    def __init__(self, root_dir: str, job_info, blueprint: quixote.Blueprint):
        self.root_dir = root_dir
        self.job_info = job_info
        self.blueprint = blueprint

    def fetch_data(self) -> ExecutableJobHandle:
        """
        Fetch the data on which this job will operate

        The fetcher functions specified in the blueprint will be used
        """
        delivery_path = os.path.join(self.root_dir, "rendu")
        resources_path = os.path.join(self.root_dir, "resources")
        with quixote.new_context(
                **get_config(),
                **self.job_info,
                delivery_path=delivery_path,
                resources_path=resources_path
        ):
            for fetcher in self.blueprint.fetchers:
                fetcher()
        return ExecutableJobHandle(self.root_dir, self.job_info, self.blueprint)


class JobWorkspaceHandle:
    """
    Class representing a job workspace, that is, a job with its working directory and its resources
    """

    def __init__(self, root_dir: str, job_info, blueprint: quixote.Blueprint):
        self.root_dir = root_dir
        self.job_info = job_info
        self.blueprint = blueprint

    def build_job_environment(self) -> DataAwaitingJobHandle:
        """
        Build the environment specified by the blueprint for this job.

        The builder functions specified in the blueprint will be used to build a suitable Docker image
        """
        env_dir = os.path.join(self.root_dir, "env")
        try:
            os.mkdir(env_dir)
            generator = panza.build.create_script_generator(GeneratorType.DOCKER)
            generator.generate_to_file(self.blueprint, os.path.join(env_dir, "Dockerfile"), extra_context=get_config())
            os.system(f"docker build -t test-{self.job_info['module_id']}-{self.job_info['activity_id']} {env_dir}")
            shutil.rmtree(env_dir)
        except Exception as e:
            raise EnvironmentBuildError(e)
        return DataAwaitingJobHandle(self.root_dir, self.job_info, self.blueprint)


class JobInfo:
    """
    Class representing a job information, that is, a directory containing the job's resources, and its metadata
    """

    def __init__(self, job_dir: str, job_info):
        self.job_dir = job_dir
        self.job_info = job_info

    def create_workspace(self) -> JobWorkspaceHandle:
        """
        Create a workspace in which the job can be executed.

        This will create a directory, copy the required files, and load the blueprint
        """
        job_name = f"{self.job_info['module_id']}-{self.job_info['activity_id']}-{self.job_info['group_id']}"
        root_dir = os.path.join(get_config()["root_dir"], job_name)

        try:
            shutil.copytree(self.job_dir, root_dir)
        except OSError as e:
            raise WorkspaceCreationError(e)

        blueprint = panza.BlueprintLoader.load_from_directory(root_dir)

        return JobWorkspaceHandle(root_dir, self.job_info, blueprint)
